@@ -48,12 +48,11 @@ do
    end
 end
 
+local current_dir_with_cache
 do
    local cache_pwd
-   --- Obtain current directory.
-   -- Uses the module's internal directory stack.
-   -- @return string: the absolute pathname of the current directory.
-   function tools.current_dir()
+
+   current_dir_with_cache = function()
       local current = cache_pwd
       if not current then
          local pipe = io.popen(fs.quiet_stderr(vars.PWD))
@@ -64,7 +63,14 @@ do
       for _, directory in ipairs(dir_stack) do
          current = fs.absolute_name(directory, current)
       end
-      return current
+      return current, cache_pwd
+   end
+
+   --- Obtain current directory.
+   -- Uses the module's internal directory stack.
+   -- @return string: the absolute pathname of the current directory.
+   function tools.current_dir()
+      return (current_dir_with_cache()) -- drop second return
    end
 end
 
@@ -107,9 +113,11 @@ end
 -- @return boolean: true if command succeeds (status code 0), false
 -- otherwise.
 function tools.execute_string(cmd)
-   local current = fs.current_dir()
+   local current, cache_pwd = current_dir_with_cache()
    if not current then return false end
-   cmd = fs.command_at(current, cmd)
+   if current ~= cache_pwd then
+      cmd = fs.command_at(current, cmd)
+   end
    local code = os.execute(cmd)
    if code == 0 or code == true then
       return true
@@ -140,8 +148,8 @@ end
 -- filename can be given explicitly as this second argument.
 -- @param cache boolean: compare remote timestamps via HTTP HEAD prior to
 -- re-downloading the file.
--- @return (boolean, string): true and the filename on success,
--- false and the error message on failure.
+-- @return (boolean, string, string): true and the filename on success,
+-- false and the error message and code on failure.
 function tools.use_downloader(url, filename, cache)
    assert(type(url) == "string")
    assert(type(filename) == "string" or not filename)
@@ -150,7 +158,7 @@ function tools.use_downloader(url, filename, cache)
 
    local downloader, err = fs.which_tool("downloader")
    if not downloader then
-      return nil, err
+      return nil, err, "downloader"
    end
 
    local ok = false
@@ -185,7 +193,7 @@ function tools.use_downloader(url, filename, cache)
       return true, filename
    else
       os.remove(filename)
-      return false, "failed downloading " .. url
+      return false, "failed downloading " .. url, "network"
    end
 end
 
